from enum import Enum

class TokenEnum(Enum):
  ATE = 'até'
  ATR = '='
  COLCHETEA = '['
  COLCHETEF = ']'
  CHAVEA = '{'
  CHAVEF = '}'
  DE = 'de'
  E = 'e'
  FUNCAO = 'funcao'
  RETORNE = 'retorne'
  FIMFUNCAO = 'fim_funcao'
  ENQUANTO = 'enquanto'
  ENTAO = 'então'
  ESCREVA = 'escreva'
  FACA = 'faça'
  FIMPARA = 'fim_para'
  FIMSE = 'fim_se'
  FIMENQUANTO = 'fim_enquanto'
  ID = 'identifier'
  LEIA = 'leia'
  LOGDIFF = '<>'
  LOGIGUAL = '='
  LOGMAIOR = '>'
  LOGMAIORIGUAL = '>='
  LOGMENOR = '<'
  LOGMENORIGUAL = '<='
  NAO = 'não'
  NUMINT = 'integer value'
  OPDIVI = '/'
  OPMAIS = '+'
  OPMENOS = '-'
  OPMULTI = '*'
  OU = 'ou'
  PARA = 'para'
  PARAB = '('
  PARFE = ')'
  PASSO = 'passo'
  SE = 'se'
  SENAO = 'senão'
  STRING = 'string value'
  TIPO = 'inteiro'
  ALGORITMO = 'algoritmo'
  VAR = 'var'
  COMMA = ','
  COLON = ':'
  PONTO = '.'
  INICIO = 'inicio'
  FIMALGORITMO = 'fim_algoritmo'
  USAR = 'usar'
  END_OF_FILE = '__EOF__'
  CONTINUE = 'continue'
  INTERROMPA = 'interrompa'
  CLASSE = 'classe'
  METODO = 'metodo'
  NOVO = 'novo'
  FIMCLASSE = 'fim_classe'
